#
#  FILE NAME:
#      dictree_helpers.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import logging

# P3 imports:
from p3.logic.error import LogicModuleError


_LOG = logging.getLogger( "p3.logic" )


def _adapt_path( path ):
    if isinstance( path, basestring ):
        if path:
            return path.split(".")
        else:
            return []
    else:
        return list(path)


def dictree_get_opt( dictree, curr_parts, prev_parts=[], default=None ):
    try:
        return _dictree_get_req( dictree, _adapt_path(curr_parts), _adapt_path(prev_parts) )
    except LogicModuleError as e:
        return default


def dictree_get_req( dictree, curr_parts, prev_parts=[] ):
    return _dictree_get_req( dictree, _adapt_path(curr_parts), _adapt_path(prev_parts) )


def _dictree_get_req( dictree, curr_parts, prev_parts ):
    if not curr_parts:
        return dictree

    curr_part = curr_parts[0]
    next_parts = curr_parts[1:]

    try:
        curr_entry = dictree[curr_part]
    except KeyError as e:
        raise LogicModuleError(
            LogicModuleError.codes.dictree_path_get_failed,
            "Couldn't find the '{0:s}' dictree path."
            "".format( '.'.join( prev_parts + [curr_part] ) ) )
    except TypeError as e:
        raise LogicModuleError(
            LogicModuleError.codes.dictree_path_get_failed,
            "Couldn't subscript the '{0:s}' dictree item (it is not dictionary like)."
            "".format( '.'.join( prev_parts ) ) )

    if next_parts:
        return _dictree_get_req( curr_entry, next_parts, prev_parts + [curr_part] )
    else:
        return curr_entry


def dictree_set( dictree, curr_parts, value, prev_parts=[], force_type_change=False ):
    _dictree_set(
        dictree, _adapt_path(curr_parts), value, _adapt_path(prev_parts), force_type_change )


def _dictree_set( dictree, curr_parts, value, prev_parts, force_type_change ):
    assert curr_parts

    curr_part = curr_parts[0]
    next_parts = curr_parts[1:]

    if next_parts:
        try:
            curr_entry = dictree.setdefault( curr_part, {} )
            _dictree_set( curr_entry, next_parts, value, prev_parts + [curr_part], force_type_change )
        except AttributeError as e:
            raise LogicModuleError(
                LogicModuleError.codes.dictree_path_set_failed,
                "Couldn't subscript the '{0:s}' dictree item (it is not dictionary like)."
                "".format( '.'.join( prev_parts + [curr_part] ) ) )
    else:
        old_value = dictree.get( curr_part )

        if old_value is None or isinstance(value, type(old_value)) or force_type_change:
            dictree[curr_part] = value
        else:
            raise LogicModuleError(
                LogicModuleError.codes.dictree_path_set_failed,
                "Couldn't set the '{0:s}' dictree item as its new type ({1:s}) differs from the"
                " existing one ({2:s})."
                "".format(
                    '.'.join(prev_parts + [curr_part]), type(value).__name__,
                    type(old_value).__name__ ) )


def dictree_del( dictree, curr_parts, prev_parts=[] ):
    _dictree_del( dictree, _adapt_path(curr_parts), _adapt_path(prev_parts) )

def _dictree_del( dictree, curr_parts, prev_parts ):
    assert curr_parts

    last_part = curr_parts[-1:][0]
    get_parts = curr_parts[:-1]

    if get_parts:
        try:
            parent = _dictree_get_req( dictree, get_parts, prev_parts )
        except LogicModuleError as e:
            if e.error_code == LogicModuleError.codes.dictree_path_get_failed:
                raise LogicModuleError(
                    LogicModuleError.codes.dictree_path_del_failed,
                    e.error_message )
    else:
        parent = dictree

    try:
        del parent[last_part]
    except KeyError as e:
        raise LogicModuleError(
            LogicModuleError.codes.dictree_path_del_failed,
            "Couldn't find the '{0:s}' dictree path."
            "".format( '.'.join( prev_parts + curr_parts ) ) )
    except TypeError as e:
        raise LogicModuleError(
            LogicModuleError.codes.dictree_path_del_failed,
            "Couldn't subscript the '{0:s}' dictree item (it is not dictionary like)."
            "".format( '.'.join( prev_parts + get_parts ) ) )
